/*
* This file is part of Firestorm NIDS
* Copyright (c) 2002 Gianni Tedesco
* This program is released under the terms of the GNU GPL version 2
*
* This code handles alerts, it mainly just calls functions in
* elog_write.c.
*
* Functions:
*  o alert_hup() - rotate log on demand
*  o cb_dir() - set log spool directory path
*  o cb_stormwall() - set stormwall mode
*  o alert_conf_hook() - called for each output directive
*  o alert_conf_go() - open all log spools
*  o alert_ratelimit() - token bucket filter
*  o alert_do_ratelimit() - wrapper to do the correct ratelimit
*  o alert() - exported API, do an alert
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <fcntl.h>
#include <netinet/in.h>

#include <firestorm.h>
#include <packet.h>
#include <alert.h>
#include <elog.h>
#include <elog_write.h>
#include <signature.h>
#include <decode.h>
#include <cleanup.h>
#include <sensor.h>
#include <args.h>
#include <capture.h>
#include <target.h>

/* By sending a SIGHUP firestorm can cause
 * a log rotation at any time */
void alert_hup(void)
{
	spool_rotate_all();
}

int cb_dir(struct arg *a, void *priv)
{
	struct elog_spool *s=priv;
	static char sw[]=".stormwall";
	int l;

	if ( !s )
		return 0;

	if ( !a->val.v_str )
		return 0;

	if ( *a->val.v_str==0 )
		return 0;

	if ( s->rotate_fn )
		return 0;

	l=strlen(a->val.v_str);

	/* We add a slash unconditionally */
	if ( a->val.v_str[l-1]=='/' ) {
		a->val.v_str[l-1]=0;
		l-=1;
	}

	/* Allocate for the log dir */
	if ( !(s->log_dir=malloc(l + ALERT_FNLEN + 2)) ) {
		return 0;
	}

	s->rotate_fn=s->log_dir + l + 1;
	sprintf(s->log_dir, "%s/", a->val.v_str);

	/* Allocate enough space for the whole filename */
	if ( !(s->alert_fn=malloc(l + strlen(ALERT_FN) + 2)) ) {
		free(s->log_dir);
		return 0;
	}

	/* Figure out stormwall fifo filename */
	if ( !(s->fifo_fn=malloc(l+sizeof(sw)+2)) ) {
		free(s->alert_fn);
		free(s->log_dir);
		return 0;
	}

	/* Generate the filenames */
	sprintf(s->alert_fn, "%s/%s", a->val.v_str, ALERT_FN);
	sprintf(s->fifo_fn, "%s/%s", a->val.v_str, sw);

	return 1;
}

int cb_stormwall(struct arg *a, void *priv)
{
	struct elog_spool *s=priv;

	if ( !s )
		return 0;

	if ( !strcmp("none", a->val.v_str) ) {
		s->stormwall=STORMWALL_NONE;
	}else if ( !strcmp("wait", a->val.v_str) ) {
		s->stormwall=STORMWALL_WAIT;
	}else if ( !strcmp("fail", a->val.v_str) ) {
		s->stormwall=STORMWALL_FAIL;
	}else{
		mesg(M_ERR, "output: stormwall can be one of 'none',"
			"'wait' or 'fail'");
		return 0;
	}

	return 1;
}

/* called once for each 'output' directive on the config file */
void alert_conf_hook(char *args)
{
	struct elog_spool *s;
	size_t buf_sz=SPOOL_DEFAULT_BUFFER;
	static struct arg a[]={
		{"dir", ARGTYPE_STRING, cb_dir},
		{"stormwall", ARGTYPE_STRING, cb_stormwall},
		{"size", ARGTYPE_PBYTES, NULL},
		{"minutes", ARGTYPE_PUINT, NULL},
		{"buf", ARGTYPE_PBYTES, NULL},
		{NULL, ARGTYPE_NOP, NULL}
	};

	/* Allocate the spool */
	if ( !(s=spool_new()) )
		cleanup(EXIT_ERR, "alert: spool_new(): %s", get_err());

	/* setup the arg pointers */
	a[2].val.vp_bytes=&s->max_bytes;
	a[3].val.vp_uint=&s->max_time;
	a[4].val.vp_bytes=&buf_sz;

	/* parse the damn things */
	if ( args ) {
		switch ( args_parse(a, args, s) ) {
		case -1:
			mesg(M_ERR, "output: parse error: %s", args);
		case 0:
			cleanup(EXIT_ERR,"output: failed");
		default:
			break;
		}
	}

	/* Must specify a log dir */
	if ( !s->alert_fn )
		cleanup(EXIT_ERR, "output: no dir specified: %s", args);

	/* Need to allocate before before continuing */
	if ( !spool_set_buf(s, buf_sz) )
		cleanup(EXIT_ERR, "output: unable to allocate buffer");

	/* purely informational */
	if ( s->max_bytes )
		mesg(M_INFO,"alert: %s: max log size: %uKB",
			s->log_dir, s->max_bytes/1024);

	/* use minutes granularity */
	if ( s->max_time ) {
		s->max_time*=60;
		mesg(M_INFO,"alert: %s: max log age: %u hrs %u mins",
			s->log_dir,
			s->max_time/(60*60),
			(s->max_time%(60*60))/60);
	}

	/* say whether buffered or not */
	mesg(M_INFO, "alert: %s: %sbuffered output: %uKB buffer",
		s->log_dir,
		buf_sz ? "" : "non-",
		buf_sz/1024);

	/* ok, all done, add to the list */
	spool_link(s);
}

/* Actually open files etc.. */
void alert_conf_go(void)
{
	spool_init_all();
}

/* Token bucket rate-limiting filter. Return values:
 * returns -1 if events are coming in too fast
 * returns zero if rate is normal
 * returns n>0 if rate is normal but some were dropped */
static inline int
alert_ratelimit(struct packet *pkt, struct tokenbucket *a)
{
	unsigned long h;
	int ret;

	if ( !a->cost ) return 0;

	/* Build a single time value with 0.01s resolution */
	h=pkt->time.tv_sec * RATE_SEC;
	h+=pkt->time.tv_usec / (1000000UL/RATE_SEC);

	a->toks += h - a->last_msg;
	a->last_msg=h;

	if ( a->toks > a->burst )
		a->toks=a->burst;

	if ( a->toks >= a->cost ) {
		ret=a->missed;
		a->missed=0;
		a->toks -= a->cost;
		return ret;
	}

	a->missed++;
	return -1;
}

/* return zero for OK, -1 to suppress */
static inline int
alert_do_ratelimit(struct tokenbucket *tb,
			struct packet *pkt,
			char *s1, char *s2)
{
	int missed;

	if ( !tb )
		return 0;

	switch ( (missed=alert_ratelimit(pkt,tb)) ) {
	case -1:
		return -1;
	case 0:
		return 0;
	default:
		mesg(M_DEBUG, "alert: [%s] %s: %i alerts suppressed",
			s1, s2, missed);
		return 0;
	}
}

/* This is where alerts come to after rate-limiting */
void alert(struct alert *a, struct packet *pkt)
{
	if ( !pkt )
		return;

	/* Token bucket filters are hierarchical */
	if ( alert_do_ratelimit(&a->t, pkt, "sig", a->alert)<0 )
		return;

	if ( alert_do_ratelimit(a->gen->t, pkt, "generator",
				a->gen->name)<0 )
		return;

#if 0
	mesg(M_INFO, "%s", a->alert);
#endif

	/* write the packet */
	spool_packet(a, pkt);
}
