/*
* This file is part of tcpidx
* Copyright (c) 2003 Gianni Tedesco
* Released under the terms of the GNU GPL version 2
*
* VERY fast output buffering for files.
*/

#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>

#include <fbuf.h>

/* fbuf_attach:
 * @b: an fbuf structure to use
 * @fd: file descriptor to write to
 * @bufsz: size of output buffer
 *
 * Attach a file descriptor to a buffer and ready it for use.
 *
 * Return value: zero on success, less than zero for error.
 */
int fbuf_attach(struct fbuf *b, int fd, size_t bufsz)
{
	if ( !bufsz )
		bufsz = 8192;

	b->buf = malloc( bufsz );
	if ( !b->buf )
		return -1;

	b->fd = fd;
	b->ptr = b->buf;
	b->buf_len = bufsz;
	b->buf_sz = bufsz;

	return 0;
}

/* fbuf_flush:
 * @b: the fbuf structure to flush
 *
 * Flush the userspace buffer to disk. Note this does not call fsync()
 * so do not rely on it in order to verify that data is written to disk.
 *
 * Return value: zero on success, less than zero for error.
 */
#include <stdio.h>
int fbuf_flush(struct fbuf *b)
{
	int len = b->buf_sz - b->buf_len;
	int ret;

	/* buffer empty */
	if ( len <= 0 )
		return 0;

again:
	ret=write(b->fd, b->buf, len);
	if ( ret == len ) {
		b->ptr = b->buf;
		b->buf_len = b->buf_sz;
		return 0;
	}

	if ( ret < 0 && errno==EINTR )
		goto again;

	return -1;
}

/* fbuf_detach:
 * @b: the fbuf structure to finish up with.
 *
 * Flush the buffer and free up any allocated memory. Note that this
 * does not close the file descriptor, you must do that yourself.
 *
 * Return value: zero on success, less than zero for error.
 */
int fbuf_detach(struct fbuf *b)
{
	int ret=0;

	if ( fbuf_flush(b)<0 )
		ret = -1;

	if ( b->buf )
		free(b->buf);
	b->buf = NULL;
	b->ptr = NULL;
	b->buf_len = 0;

	return ret;
}

/* fbuf_close:
 * @b: the fbuf structure to finish up with.
 *
 * Like fbuf_detach except that we close the file descriptor for
 * writing and we make sure that the data is written safely to disk.
 * before returning.
 *
 * Return value: zero on success, less than zero for error.
 */
int fbuf_close(struct fbuf *b)
{
	int ret=0;

	if ( b->fd == -1 )
		goto noclose;

	if ( fbuf_flush(b)<0 )
		ret = -1;

	/* don't error if the output file is a special file
	 * which does not support fsync (eg: a pipe)
	 */
	if ( fsync(b->fd) && errno!=EROFS && errno!=EINVAL )
		ret = -1;

	if ( close(b->fd) )
		ret = -1;

noclose:
	if ( b->buf )
		free(b->buf);
	b->buf = NULL;
	b->ptr = NULL;
	b->buf_len = 0;

	return ret;
}

/* fbuf_newfd:
 * @b: the fbuf structure to modify
 * @b: the new file descriptor
 *
 * Flush current file safely to disk and start writing to a new FD.
 *
 * Return value: zero on success, less than zero for error.
 */
int fbuf_newfd(struct fbuf *b, int fd)
{
	int ret=0;

	if ( b->fd == -1 )
		goto noflush;

	if ( fbuf_flush(b)<0 )
		ret = -1;

	/* don't error if the output file is a special file
	 * which does not support fsync (eg: a pipe)
	 */
	if ( fsync(b->fd) && errno!=EROFS && errno!=EINVAL )
		ret = -1;

	if ( close(b->fd) )
		ret = -1;

noflush:
	b->fd = fd;

	return ret;
}

/* fbuf_write_slow:
 * @b: an fbuf structure to use
 * @buf: pointer to the data you want to write
 * @len: size of data pointed to by @buf
 *
 * Slow path for writing to the buffer, do not call directly instead
 * use fbuf_write().
 *
 * Return value: zero on success, less than zero for error.
 */
int fbuf_write_slow(struct fbuf *b, void *buf, size_t len)
{
	/* fill up the buffer before flushing, we already know
	 * that len >= b->buf_len so a full buffer flush is
	 * inevitable.
	 */
	memcpy(b->ptr, buf, b->buf_len);
	buf += b->buf_len;
	len -= b->buf_len;
	b->ptr += b->buf_len;
	b->buf_len = 0;
	if ( fbuf_flush(b) < 0 )
		return -1;

	/* If the remaining data is the same size as the buffer then
	 * buffering doing an un-necessary copy of the data.
	 *
	 * If the remaining data is bigger than the buffer then we
	 * must write it out right away */
	if ( len >= b->buf_sz ) {
		int ret;
again:
		ret = write(b->fd, buf, len);
		if ( ret == len )
			return 0;

		if ( ret < 0 && errno==EINTR )
			goto again;

		return -1;
	}

	/* normal write - len may be zero */
	memcpy(b->ptr, buf, len);
	b->ptr += len;
	b->buf_len -= len;

	return 0;
}
