/*
* This file is part of Firestorm NIDS
* Copyright (c) 2002 Gianni Tedesco
* This program is released under the terms of the GNU GPL version 2
*/
#include <config.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <stdarg.h>
#include <unistd.h>
#ifdef HAVE_GETOPT_H
#include <getopt.h>
#endif
#include <fcntl.h>
#include <string.h>
#include <netinet/in.h>

#include <firestorm.h>
#include <packet.h>
#include <cleanup.h>
#include <loader.h>
#include <alert.h>
#include <args.h>
#include <target.h>
#include <signature.h>
#include <decode.h>
#include <matcher.h>
#include <detect.h>
#include <capture.h>
#include <elog.h>
#include <elog_read.h>
#include <target.h>

#include <mpool.h>
#include <fbuf.h>

#include <firecat.h>

#ifdef HAVE_GETOPT_LONG
static const struct option opts[]={
	{"plugin-dir", 	1, 0, 'p'},
	{"options", 	1, 0, 'o'},
	{"format", 	1, 0, 'f'},
	{"index", 	0, 0, 'i'},
	{"query", 	1, 0, 'q'},
	{"help", 	0, 0, 'h'},
	{NULL, 0, 0, 0}
};
#endif

/* Store argv[0] */
static char *cmd;

/* for MODE_CAT */
static struct target *t;
static void *t_priv;
static char *oformat;
static char *p_options;
static char *query;

/* Functions we may need to fake */
int return_one(void) {return 1;}
void return_void(void) {return;}
void *return_null(void) {return NULL;}
void serial_number(serial_t *s){*s=0;}

struct {
	const char *name;
	void *data;
}exports[]={
	/* Generic */
	{"mesg",		mesg},
	{"args.parse", 		args_parse},
	{"serial_number",	serial_number},

	/* Object lookup */
	{"decode.subproto", 	decode_subproto},
	{"decode.proto", 	decode_proto},
	{"matcher.find",	return_null},

	/* Needed for signature engines */
	{"alert",		return_void},
	{"generator.add",	return_one},
	{"detect.add_sig",	return_one},
	{"detect.optimize",	return_void},
	{"detect.free_sig",	return_void},
	{"detect",		return_void},

	{NULL,NULL}
};

/* Only print wanings,errors,and critical messages */
void mesg(unsigned char code, char *fmt, ...)
{
	va_list va;

	if ( code > M_WARN )
		return;

	/* Use stderr cos output may use stdout */
	va_start(va, fmt);
	fprintf(stderr, "%s: ", cmd);
	vfprintf(stderr, fmt, va);
	fprintf(stderr, "\n");
	va_end(va);
}

void *plugin_callback(const char *item)
{
	u_int32_t i;

	for(i=0; exports[i].name; i++) {
		if ( !strcmp(exports[i].name, item) ) {
			return exports[i].data;
		}
	}

	fprintf(stderr, "resource: request for unknown object: '%s'\n", item);
	return NULL;
}

void get_plugins(void)
{
	/* Load target plugins */
#if 1
	loader_load_dir(PLUGIN_DIR "/protocols");
	loader_load_dir(PLUGIN_DIR "/output");
#else
	loader_load_dir("./decode_plugins/.libs");
	loader_load_dir("./target_plugins/.libs");
#endif

	decode_load();
	target_load();
}

/* Display usage message and quit */
void usage(int ecode)
{
	extern struct target *targets;
	struct target *t;

	fprintf(stderr, "Usage: %s [-f format] [-o \"options\"] [FILE]...\n", cmd);
	fprintf(stderr, "Concatenates and converts firestorm extended log (elog) files.\n\n");
	fprintf(stderr, "    --format, -f      Choose a coversion format\n");
	fprintf(stderr, "    --options, -o     Plugin specific option string\n");
	fprintf(stderr, "    --plugin-dir, -p  Load extra plugins from this directory\n");
	fprintf(stderr, "    --index, -i       Create query indexes\n");
	fprintf(stderr, "    --query, -q       Select only certain packets\n");
	fprintf(stderr, "    --help, -h        This message\n\n");
	fprintf(stderr, "With no FILE, or when FILE is -, read standard input.\n\n");
	fprintf(stderr, "Loading plugins from: %s\n", PLUGIN_DIR "/protocols");
	fprintf(stderr, "Loading plugins from: %s\n\n", PLUGIN_DIR "/output");
	fprintf(stderr, "Available conversion formats:\n");
	fprintf(stderr, "    elog - no conversion\n");

	/* Display available target plugins */
	get_plugins();
	for(t=targets; t; t=t->next) {
		fprintf(stderr, "    %s - %s\n", t->name, t->desc);
	}

	fprintf(stderr, "\nReport bugs to <gianni@scaramanga.co.uk>\n");
	cleanup(ecode,NULL);
}

int cbfn_target(struct elog_pkt *p, void *data)
{
	return elog_pkt_alert(p, t->target, data);
}

/* Process a file with a target plugin */
void cat_file(char *fn, int fd)
{
	void *ep;

	if ( !(ep=elog_new(fd)) ) {
		mesg(M_ERR, "%s: elog_new(): %s\n", fn, get_err());
		return;
	}

	elog_set_alert_fn(ep, cbfn_target);

	if ( elog_run(ep, t_priv)<=0 ) {
		elog_free(ep);
		cleanup(EXIT_ERR, NULL);
	}

	elog_free(ep);
}

/* Basically we just load and initialise the required target plugin */
void cat_init(void)
{
	get_plugins();

	/* Use elog if none specified */
	if ( !oformat )
		oformat="elog";

	/* Get the right output plugin */
	if ( !(t=target_find(oformat)) ) {
		mesg(M_ERR, "%s: Cannot find specified output format",
			oformat);
		cleanup(EXIT_ERR,NULL);
	}

	/* Initialise it with user specified arguments*/
	if ( !t->validate(p_options, &t_priv) ) {
		mesg(M_ERR, "%s: Invalid options for output plugin",
			oformat);
		cleanup(EXIT_ERR,NULL);
	}
}

void cat_fini(void)
{
	if ( !t->cleanup(t_priv) )
		cleanup(EXIT_ERR,NULL);
}

/* TODO */
void idx_file(char *fn, int fd)
{
}

/* TODO */
void query_file(char *fn, int fd)
{
}

/* Define the various modes of operation */
static struct firecat_mode mode[]={
	[ MODE_CAT ] { .init=cat_init, .file=cat_file, .fini=cat_fini },
	[ MODE_IDX ] { .init=get_plugins, .file=idx_file, .fini=NULL },
	[ MODE_QER ] { .init=cat_init, .file=query_file, .fini=cat_fini },
};

int main(int argc, char **argv)
{
	int m=MODE_CAT;
	int c, opti;
	int fd;
	int i;

	/* Initialise what we need to */
	cmd=argv[0];
	loader_init(plugin_callback);

	/* Do command line args */
#ifdef HAVE_GETOPT_LONG
	while( (c=getopt_long(argc, argv, "p:q:if:o:h?",
		opts, &opti))!=-1 )
#else
	while( (c=getopt(argc, argv, "p:q:if:o:h?"))!=-1 )
#endif
	{
		switch (c) {
		case 'i':
			m=MODE_IDX;
			break;
		case 'q':
			m=MODE_QER;
			query=optarg;
			break;
		case 'p':
			loader_load_dir(optarg);
			break;
		case 'f':
			if ( oformat )
				usage(EXIT_ERR);
			oformat=optarg;
			break;
		case 'o':
			if ( p_options )
				usage(EXIT_ERR);
			p_options=optarg;
			break;
		case 'h':
		case '?':
			usage(EXIT_OK);
			break;
		default:
			usage(EXIT_ERR);
		};
	}

	/* TODO: Check for conflicting parameters */

	if ( mode[m].init )
		mode[m].init();

	/* cat semantics. if no file specified, or file is -, read
	 * standard input */
	if ( optind >= argc ) {
		mode[m].file("(stdin)", STDIN_FILENO);
	}

	for(i=optind; i<argc; i++ ) {
		/* use stdin for - */
		if ( !strcmp(argv[i],"-") ) {
			mode[m].file("(stdin)", STDIN_FILENO);
		}else{
			/* Open the input file */
			if ( (fd=open(argv[i],O_RDONLY))<0 ) {
				mesg(M_ERR, "%s: open(): %s",
					argv[i], get_err());
				cleanup(EXIT_ERR,NULL);
			}

			mode[m].file(argv[i], fd);
			close(fd);
		}
	}

	if ( mode[m].fini )
		mode[m].fini();

	cleanup(EXIT_OK,NULL);
	return 0;
}
